#ifndef __BOARD_H__
#define __BOARD_H__

#include "pattern.h"

#define TRUE  1
#define FALSE 0

enum STONE {
	S_EMPTY,	// 空点
	S_BLACK,	// 黒　
	S_WHITE,	// 白
	S_OB,		// 盤外
	S_MAX,
};

#define OTHER_COLOR( color ) ((color)^0x3)	// 色反転 

#define PASS   0
#define RESIGN -1

#define KOMI 6.5	// コミ

#define PURE_BOARD_SIZE	9	// 盤の本当の大きさ
#define OB_SIZE 1			// 盤外の大きさ
#define BOARD_SIZE	(PURE_BOARD_SIZE + (OB_SIZE + OB_SIZE))	// 盤の大きさ 

#define BOARD_START OB_SIZE
#define BOARD_END	(OB_SIZE+PURE_BOARD_SIZE-1)

#define PURE_BOARD_MAX	(PURE_BOARD_SIZE * PURE_BOARD_SIZE)
#define BOARD_MAX		(BOARD_SIZE * BOARD_SIZE)

#define MAX_STRINGS (PURE_BOARD_SIZE * PURE_BOARD_SIZE * 4/5)	// 連の最大数 

#define BOARD_FIRST	POS(BOARD_START, BOARD_START)
#define BOARD_LAST	POS(BOARD_END, BOARD_END)

#define MAX_LIBERTIES 2

#define MAX_RECORDS 1000
#define MAX_MOVES MAX_RECORDS-1	// 最大手数 

#define POS(x, y)     ((x) + (y) * BOARD_SIZE)
#define X(pos)        ((pos) % BOARD_SIZE)
#define Y(pos)        ((pos) / BOARD_SIZE)
#define NORTH(pos)    ((pos) - BOARD_SIZE)
#define EAST(pos)     ((pos) + 1)
#define SOUTH(pos)    ((pos) + BOARD_SIZE)
#define WEST(pos)     ((pos) - 1)

// 近傍の座標 
#define NE  (-BOARD_SIZE+1)
#define NW  (-BOARD_SIZE-1)
#define SE  (BOARD_SIZE+1)
#define SW  (BOARD_SIZE-1)


///***			構造体				***///

// 着手 
struct move_data {
	int color;
	int pos;
};

// 石 
struct stone_data {
	int color;
	int string_number;
	int next_stone;
};

// 連 
struct string_data {
	int color;
	int size;
	int origin;
	int liberties;
	int mark;
};

// 対局の全情報 
struct game {
	int prisoner[S_MAX]; // アゲハマ

	int moves;				
	struct move_data record[MAX_RECORDS];

	struct move_data ko; // 劫
	int ko_move;

	struct stone_data board[BOARD_MAX];

	struct string_data string[MAX_STRINGS];
	int next_string;
	int liberty;
	int liberty_mark;
	int string_mark;
	int mark[BOARD_MAX];

	int pat3[BOARD_MAX];

	int candidates[BOARD_MAX]; // 候補手
	int sum_candidates_column[BOARD_SIZE];
	int sum_candidates;
};

///***			変数宣言			***///

// X座標 
extern int board_x[BOARD_MAX];

// Y座標 
extern int board_y[BOARD_MAX];

// 4近傍の数 
extern int nb4_empty[PATTERN3_MAX];
extern int nb4_black[PATTERN3_MAX];
extern int nb4_white[PATTERN3_MAX];

// 眼 
extern int eye[PATTERN3_MAX];

///***			関数宣言			***///

// 初期化 
struct game *InitializeGame( void );
struct game *SetUpGame( void );

// 初期設定 
void ClearGame( struct game *game );

// コピー 
void CopyGame( struct game *g2, struct game *g1 );

// メモリ解放 
void DoneGame( struct game *game );

// 定数の設定 
void BoardConst( void );

// 近傍の数の初期化 
void InitializeNeighbor( void );

// 眼の初期化 
void InitializeEye( void );

// 碁盤を表示 
void DisplayBoard( struct game *game  );

// 合法手の判定
int SimpleCheckLegal( struct game *game, int color, int pos );
int CheckLegal( struct game *game, int color, int pos );

// 眼の判定 
int CheckEye( struct game *game, int color, int pos );

// 合法手かつ眼でない 
int CheckLegalAndEye( struct game *game, int color, int pos );

// 自殺手の判定 
int CheckSuicide( struct game *game, int color, int pos );

// 碁盤に石を置く
void SetStone( struct game *game, int color, int pos );

// 囲まれた石を取り除く 
int	RemoveStone( struct game *game, int str );

// 連の新規作成 
void CreateNewString( struct game *game, int pos );

// 周囲の連に石を追加 
void ExtendNeighborString( struct game *game, int str, int pos );

// 連の作成 + 同化 
void AssimilateNeighborString( struct game *game, int pos );

// 連の同化 
void AssimilateString( struct game *game, int str, int pos );

// 連の表示 
void DisplayString( struct game *game );

// ダメの更新 
void UpdateLiberties( struct game *game, int str );

// 勝敗の判定 
int BoardScore( struct game *game );


#endif